"""
Non thread-safe object which provides all the RSE/PNN information
required for automatic data placement.
It uses Rucio for checking quota available and data usage
"""
from __future__ import division, print_function
from builtins import str as newstr, bytes, object
from future.utils import viewitems

from WMCore.MicroService.Tools.Common import getMSLogger, gigaBytes, teraBytes


class RSEQuotas(object):
    """
    Class which represents a list of RSEs, their quota and
    their storage usage
    """

    def __init__(self, dataAcct, quotaFraction, **kwargs):
        """
        Executes a basic setup, including proper logging.
        :param dataAcct: string with the Rucio account
        :param quotaFraction: float point number representing the fraction of the quota
        :param kwargs: the supported keyword arguments are:
          minimumThreshold: integer value defining the minimum available space required
          verbose: logger verbosity
          logger: logger object
        """
        self.dataAcct = dataAcct
        self.quotaFraction = quotaFraction

        self.minimumSpace = kwargs["minimumThreshold"]
        self.logger = getMSLogger(kwargs.get("verbose"), kwargs.get("logger"))
        msg = "RSEQuotas started with parameters: dataAcct=%s, quotaFraction=%s, "
        msg += "minimumThreshold=%s GB"
        self.logger.info(msg, dataAcct, quotaFraction, gigaBytes(self.minimumSpace))

        self.nodeUsage = {}
        self.availableRSEs = set()
        self.outOfSpaceNodes = set()

    def __str__(self):
        """
        Write out useful information for this object
        :return: a stringified dictionary
        """
        res = {'dataAcct': self.dataAcct, 'quotaFraction': self.quotaFraction,
               'minimumSpace': self.minimumSpace}
        return str(res)

    def getNodeUsage(self):
        """
        Return a dictionary of RSEs and a few storage statistics
        """
        return self.nodeUsage

    def getAvailableRSEs(self):
        """
        Return a list of out-of-space RSE/PNNs
        """
        return self.availableRSEs

    def getOutOfSpaceRSEs(self):
        """
        Return a list of out-of-space RSE/PNNs
        """
        return self.outOfSpaceNodes

    def fetchStorageQuota(self, dataSvcObj):
        """
        Fetch the storage quota/limit for a given Rucio account.
        :param dataSvcObj: object instance for the Rucio data service

        :return: create an instance cache structure to keep track of quota
          and available storage. The structure is as follows:
          {"pnn_name": {"quota": quota in bytes for the rucio account,
                        "bytes_limit": total space for the account/group,
                        "bytes": amount of bytes currently used/archived,
                        "bytes_remaining": space remaining for the acct/group,
                        "quota_avail": a fraction of the quota that we will use}
        """
        self.nodeUsage.clear()
        response = dataSvcObj.getAccountLimits(self.dataAcct)
        for rse, quota in viewitems(response):
            if rse.endswith("_Tape") or rse.endswith("_Export"):
                continue
            self.nodeUsage.setdefault(rse, {})
            self.nodeUsage[rse] = dict(quota=int(quota),
                                       bytes_limit=int(quota),
                                       bytes=0,
                                       bytes_remaining=int(quota),  # FIXME: always 0
                                       quota_avail=0)
        self.logger.info("Storage quota filled from Rucio")

    def fetchStorageUsage(self, dataSvcObj):
        """
        Fetch the storage usage from Rucio, which will then
        be used as part of the data placement mechanism.
        Also calculate the available quota - given the configurable quota
        fraction - and mark RSEs with less than 1TB available as NOT usable.
        :param dataSvcObj: object instance for the data service

        Keys definition is:
         * quota: the Rucio account limit
         * bytes_limit: the account quota from Rucio
         * bytes: data volume placed by Rucio
         * bytes_remaining: storage available for our account/group
         * quota_avail: space left (in bytes) that we can use for data placement
        :return: update our cache in place with up-to-date values, in the format of:
            {"pnn_name": {"bytes_limit": total space for the account/group,
                          "bytes": amount of bytes currently used/archived,
                          "bytes_remaining": space remaining for the acct/group}
        """
        self.logger.info("Using Rucio for storage usage, with acct: %s", self.dataAcct)
        for item in dataSvcObj.getAccountUsage(self.dataAcct):
            if item['rse'] not in self.nodeUsage:
                self.logger.warning("Rucio RSE: %s has data usage but no quota available.", item['rse'])
                continue
            # bytes_limit is always 0, so skip it and use whatever came from the limits call
            # bytes_remaining is always negative, so calculate it based on the limits
            quota = self.nodeUsage[item['rse']]['quota']
            self.nodeUsage[item['rse']].update({'bytes': item['bytes'],
                                                'bytes_remaining': quota - item['bytes']})

    def evaluateQuotaExceeded(self):
        """
        Goes through every single site, their quota and their remaining
        storage; and mark those with less than X TB available (1TB at the
        moment) as not eligible to receive data
        :return: updates instance structures in place
        """
        self.availableRSEs.clear()
        self.outOfSpaceNodes.clear()
        # given a configurable sub-fraction of our quota, recalculate how much storage is left
        for rse, info in viewitems(self.nodeUsage):
            quotaAvail = info['quota'] * self.quotaFraction
            info['quota_avail'] = min(quotaAvail, info['bytes_remaining'])
            if info['quota_avail'] < self.minimumSpace:
                self.outOfSpaceNodes.add(rse)
            else:
                self.availableRSEs.add(rse)
        self.logger.info("Currently %d nodes are out of space.", len(self.outOfSpaceNodes))

    def printQuotaSummary(self):
        """
        Print a summary of the current quotas, space usage and space available
        """
        self.logger.info("Summary of the current quotas in Terabytes:")
        for node in sorted(self.nodeUsage.keys()):
            msg = "  %s:\t\tbytes_limit: %.2f, bytes_used: %.2f, bytes_remaining: %.2f, "
            msg += "quota: %.2f, quota_avail: %.2f"
            self.logger.info(msg, node, teraBytes(self.nodeUsage[node]['bytes_limit']),
                             teraBytes(self.nodeUsage[node]['bytes']),
                             teraBytes(self.nodeUsage[node]['bytes_remaining']),
                             teraBytes(self.nodeUsage[node]['quota']),
                             teraBytes(self.nodeUsage[node]['quota_avail']))
        self.logger.info("List of RSE's out of quota: %s", self.outOfSpaceNodes)

    def updateNodeUsage(self, node, dataSize):
        """
        Provided a RSE/PNN name and the data size, in bytes, update the node
        storage usage by subtracting it from the current available quota.
        If it gets a list of nodes, the same dataSize is accounted for all
        of them.
        :param node: string with the PNN/RSE
        :param dataSize: integer with the amount of bytes allocated
        :return: nothing. updates nodeUsage cache
        """
        if isinstance(node, (newstr, bytes)):
            node = [node]
        if not isinstance(dataSize, int):
            self.logger.error("dataSize needs to be integer, not '%s'!", type(dataSize))
        for rse in node:
            self.nodeUsage[rse]['quota_avail'] -= dataSize



"""
For more details about StoreResults workflows, please refer to:
https://github.com/dmwm/WMCore/wiki/StoreResults-requests

Pre-requisites:
 1. a valid proxy in your X509_USER_PROXY variable
 2. wmagent env: source /data/srv/wmagent/current/apps/wmagent/etc/profile.d/init.sh
 3. have the correct permissions in CRIC, otherwise dataset migration won't work

Expected input json file like:
[{"InputDataset": "/EmbeddingRun2016B/MuTauFinalState-imputSep16DoubleMu_mirror_miniAOD-v2/USER",
  "DbsUrl": "phys03",
  "ScramArch": "slc6_amd64_gcc530",
  "SiteWhitelist": "T2_DE_DESY",
  "PhysicsGroup": "Tau POG",
  "CMSSWVersion": "CMSSW_8_0_26_patch1"},
  {..},
  {..}]

"""

import http.client
import json
import os
import sys
import time
from copy import copy

from dbs.apis.dbsClient import DbsApi

url = "cmsweb.cern.ch"
dbsApi = DbsApi(url='https://%s/dbs/prod/global/DBSMigrate/' % url)

DEFAULT_DICT = {
    "CMSSWVersion": "UPDATEME",
    "InputDataset": "UPDATEME",
    "PhysicsGroup": "UPDATEME",
    "PrepID": "UPDATEME",
    "ProcessingString": "UPDATEME",
    "RequestString": "UPDATEME",
    "ScramArch": "UPDATEME",
    "SiteWhitelist": "UPDATEME",  # Will be picked by Unified and posted again at assignment
    "AcquisitionEra": "UPDATEME",
    "ProcessingVersion": 1,  # will be updated too
    "Campaign": "StoreResults",
    "DbsUrl": "https://cmsweb-prod.cern.ch/dbs/prod/phys03/DBSReader",
    "GlobalTag": "crab3_tag",
    "Memory": 2000,
    "RequestPriority": 999999,
    "RequestType": "StoreResults",
    "SizePerEvent": 512,
    "TimePerEvent": 1}


MANDATORY_FIELDS = {"CMSSWVersion", "ScramArch", "DbsUrl",
                    "InputDataset", "SiteWhitelist", "PhysicsGroup"}


def main():
    if len(sys.argv) != 2:
        print("Usage: python injectHarvest.py <input_json_file>")
        sys.exit(0)

    inputFile = sys.argv[1]
    with open(inputFile) as fo:
        items = json.load(fo)
    # now create requests for each of the datasets
    for entry in items:
        migrateDataset(entry['InputDataset'], entry['DbsUrl'])
        newDict = buildRequest(entry)
        if newDict is None:
            # user provided incomplete data (or mistyped something)
            continue
        workflow = submitWorkflow(newDict)
        approveRequest(workflow)
    sys.exit(0)


def migrateDataset(dset, dbsInst):
    """
    Migrate dataset from the user instance to the DBS prod one.
    It returns the origin site name, which should be used for assignment
    """
    dbsInst = "https://cmsweb-prod.cern.ch/dbs/prod/%s/DBSReader" % dbsInst
    migrateArgs = {'migration_url': dbsInst, 'migration_input': dset}
    dbsApi.submitMigration(migrateArgs)
    print("Migrating dataset %s from %s to prod/global" % (dset, dbsInst))


def buildRequest(userDict):
    """
    Expects the following user data:
      CMSSWVersion, ScramArch, DbsUrl, InputDataset, SiteWhitelist and PhysicsGroup 
    """
    if set(userDict.keys()) != MANDATORY_FIELDS:
        print("ERROR: user input data is incomplete: %s" % userDict)
        print("User json *must* have these fields: %s" % MANDATORY_FIELDS)
        return None

    newSchema = copy(DEFAULT_DICT)
    newSchema.update(userDict)
    newSchema['DbsUrl'] = "https://cmsweb-prod.cern.ch/dbs/prod/%s/DBSReader" % newSchema['DbsUrl']
    # Remove spaces from the Physics Group value
    newSchema['PhysicsGroup'] = newSchema['PhysicsGroup'].replace(" ", "")
    # Set PrepID according to the date and time
    newSchema["PrepID"] = "StoreResults-%s" % time.strftime("%d%m%y-%H%M%S")
    # Truncate the ProcessingString, otherwise it can be larger than allowed
    primDset, procDset, _tier = newSchema['InputDataset'].split("/")[1:]
    acqEra, procStr = procDset.split("-", 1)
    newSchema["AcquisitionEra"] = acqEra  # should we worry about length limits?
    procStr, procVer = procStr.rsplit("-", 1)
    newSchema["ProcessingString"] = "StoreResults_" + procStr[:67]  # limit to 80 chars
    # ProcessingString cannot have a dash char
    newSchema["ProcessingString"] = newSchema["ProcessingString"].replace("-", "_")
    try:
        newSchema["ProcessingVersion"] = int(procVer[1:])
    except ValueError:
        newSchema["ProcessingVersion"] = 1
    # Use PrimaryDataset and ProcessedDataset in the RequestString
    newSchema["RequestString"] = primDset[:35] + "-" + procDset[:35]
    return newSchema


def submitWorkflow(schema):
    headers = {"Content-type": "application/json",
               "Accept": "application/json"}
    encodedParams = json.dumps(schema)
    conn = http.client.HTTPSConnection(url, cert_file=os.getenv('X509_USER_PROXY'), key_file=os.getenv('X509_USER_PROXY'))
    conn.request("POST", "/reqmgr2/data/request", encodedParams, headers)
    resp = conn.getresponse()
    data = resp.read()
    if resp.status != 200:
        print("Response status: %s\tResponse reason: %s" % (resp.status, resp.reason))
        print("Error message: %s" % resp.msg.getheader('X-Error-Detail'))
        return None
    data = json.loads(data)
    requestName = data['result'][0]['request']
    print("  Request %s successfully created.\n" % requestName)
    return requestName


def approveRequest(workflow):
    if workflow is None:
        return

    # print("Approving request...")
    encodedParams = json.dumps({"RequestStatus": "assignment-approved"})
    headers = {"Content-type": "application/json",
               "Accept": "application/json"}

    conn = http.client.HTTPSConnection(url, cert_file=os.getenv('X509_USER_PROXY'), key_file=os.getenv('X509_USER_PROXY'))
    conn.request("PUT", "/reqmgr2/data/request/%s" % workflow, encodedParams, headers)
    resp = conn.getresponse()
    data = resp.read()
    if resp.status != 200:
        print("Response status: %s\tResponse reason: %s" % (resp.status, resp.reason))
        if hasattr(resp.msg, "x-error-detail"):
            print("Error message: %s" % resp.msg["x-error-detail"])
            sys.exit(2)
    conn.close()
    # print("  Request successfully approved!")
    return


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
_File_

Data object that contains details for a single file

"""
from builtins import str, bytes
__all__ = []

from WMCore.DataStructs.Run import Run
from WMCore.DataStructs.WMObject import WMObject


class File(WMObject, dict):
    """
    _File_
    Data object that contains details for a single file

    TODO 
    - use the decorator `from functools import total_ordering` after
      dropping support for python 2.6
    - then, drop __ne__, __le__, __gt__, __ge__
    """

    def __init__(self, lfn="", size=0, events=0, checksums=None,
                 parents=None, locations=None, merged=False):
        dict.__init__(self)
        checksums = checksums or {}
        self.setdefault("lfn", lfn)
        self.setdefault("size", size)
        self.setdefault("events", events)
        self.setdefault("checksums", checksums)
        self.setdefault('runs', set())
        self.setdefault('merged', merged)
        self.setdefault('last_event', 0)
        self.setdefault('first_event', 0)
        if locations is None:
            self.setdefault("locations", set())
        else:
            self.setdefault("locations", locations)

        if parents is None:
            self.setdefault("parents", set())
        else:
            self.setdefault("parents", parents)

    def addRun(self, run):
        """
        _addRun_

        run should be an instance of WMCore.DataStructs.Run

        Add a run container to this file, tweak the run and lumi
        keys to be max run and max lumi for backwards compat.


        """

        if not isinstance(run, Run):
            msg = "addRun argument must be of type WMCore.DataStructs.Run"
            raise RuntimeError(msg)

        addFlag = False
        for runMember in self['runs']:
            if runMember.run == run.run:
                # this rely on Run object overwrite __add__ to update self
                runMember + run
                addFlag = True

        if not addFlag:
            self['runs'].add(run)
        return

    def load(self):
        """
        A DataStructs file has nothing to load from, other implementations will
        over-ride this method.
        """
        if self['id']:
            self['lfn'] = '/store/testing/%s' % self['id']

    def save(self):
        """
        A DataStructs file has nothing to save to, other implementations will
        over-ride this method.
        """
        pass

    def setLocation(self, pnn):
        # Make sure we don't add None, [], "" as file location
        if pnn:
            self['locations'] = self['locations'] | set(self.makelist(pnn))

    def __eq__(self, rhs):
        """
        File is equal if it has the same name
        """
        eq = False
        if isinstance(rhs, type(self)):
            eq = self['lfn'] == rhs['lfn']
        elif isinstance(rhs, (str, bytes)):
            eq = self['lfn'] == rhs
        return eq

    def __ne__(self, rhs):
        return not self.__eq__(rhs)

    def __hash__(self):
        thisHash = self['lfn'].__hash__()
        return thisHash

    def __lt__(self, rhs):
        """
        Sort files based on lexicographical ordering of the value connected
        to the 'lfn' key
        """
        eq = False
        if isinstance(rhs, type(self)):
            eq = self['lfn'] < rhs['lfn']
        elif isinstance(rhs, (str, bytes)):
            eq = self['lfn'] < rhs
        return eq

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def json(self, thunker=None):
        """
        _json_

        Serialize the file object.  This will convert all Sets() to lists and
        weed out the internal data structures that don't need to be shared.
        """
        fileDict = {"last_event": self["last_event"],
                    "first_event": self["first_event"],
                    "lfn": self["lfn"],
                    "locations": list(self["locations"]),
                    "id": self.get("id", None),
                    "checksums": self["checksums"],
                    "events": self["events"],
                    "merged": self["merged"],
                    "size": self["size"],
                    "runs": [],
                    "parents": []}

        for parent in self["parents"]:
            if isinstance(parent, (str, bytes)):
                # Then for some reason, we're passing strings
                # Done specifically for ErrorHandler
                fileDict['parents'].append(parent)
            elif thunker is None:
                continue
            else:
                fileDict["parents"].append(thunker._thunk(parent))

        for run in self["runs"]:
            runDict = {"run_number": run.run,
                       "lumis": run.lumis}
            fileDict["runs"].append(runDict)

        return fileDict

    def __to_json__(self, thunker=None):
        """
        __to_json__

        This is the standard way we jsonize other objects.
        Included here so we have a uniform method.
        """
        return self.json(thunker)



"""
_ContinuousSummaryHistogram_

Continuous histogram module, to be used by the TaskArchiver
to store histograms in the summary where the data is a continuous number.

Created on Nov 20, 2012

@author: dballest
"""
from __future__ import division
import math

from WMCore.DataStructs.MathStructs.SummaryHistogram import SummaryHistogram
from WMCore.Algorithms.MathAlgos import validateNumericInput
from WMCore.Algorithms.MathAlgos import calculateRunningAverageAndQValue, calculateStdDevFromQ

class ContinuousSummaryHistogram(SummaryHistogram):
    """
    _ContinuousSummaryHistogram_

    A histogram where there are continuous points
    with certain frequency, it follows
    that there is only one value in Y and
    that the average and standard deviation are
    not calculated on the frequency values but the X values.
    """

    def __init__(self, title, xLabel, yLabel = None,
                 roundingDecimals = 2, nBins = None,
                 dropOutliers = False, sigmaLimit = 5,
                 storeHistogram = True):
        """
        __init__

        Initialize a more complex histogram structure, containing different
        data to calculate online average and standard deviations. This data is also
        stored in the JSON to allow rebuilding and adding histograms.

        All histograms are binned when requested, the resolution can be specified
        through nBins, otherwise the value used is the one recommended in:
        Wand, M.P. (1997), "Data-Based Choice of Histogram Bin Width," The American Statistician, 51, 59-64.

        If specified, outlier farther than sigmaLimit standard deviations from the
        mean will not be included in the binned histogram.
        """
        # Initialize the parent object
        SummaryHistogram.__init__(self, title, xLabel)

        # Indicate this is a discrete histogram
        self.continuous = True

        # Add data only used in the continuous version
        self.yLabel            = yLabel
        self.nPoints           = 0
        self.QValue            = None
        self.average           = None

        # Configuration parameters for the continuous histograms
        self.roundingDecimals = roundingDecimals
        self.fixedNBins       = nBins
        self.dropOutliers     = dropOutliers
        self.sigmaLimit       = sigmaLimit
        self.binned           = False
        self.storeHistogram   = storeHistogram

        # Override initialization of some attributes
        self.average = 0.0
        self.stdDev  = 0.0

        return

    def addPoint(self, xValue, yLabel = None):
        """
        _addPoint_

        Add a point from a continuous set (only-numbers allowed currently) to the histogram data,
        calculate the running average and standard deviation.
        If no y-label had been specified before, one must be supplied
        otherwise the given y-label must be either None or equal
        to the stored value.
        """
        if self.binned:
            # Points can't be added to binned histograms!
            raise Exception("Points can't be added to binned histograms")

        if self.yLabel is None and yLabel is None:
            raise Exception("Some y-label must be stored for the histogram")
        elif self.yLabel is None:
            self.yLabel = yLabel
        elif yLabel is not None and self.yLabel != yLabel:
            raise Exception("Only one y-label is allowed on continuous histograms")

        if not validateNumericInput(xValue):
            # Do nothing if it is not a number
            return

        xValue = float(xValue)
        xValue = round(xValue, self.roundingDecimals)

        if self.storeHistogram:
            if xValue not in self.data:
                self.data[xValue] = 0
            self.data[xValue] += 1

        self.nPoints += 1

        (self.average, self.QValue) = calculateRunningAverageAndQValue(xValue, self.nPoints, self.average, self.QValue)

        return

    def __add__(self, other):
        #TODO: For HG1302, support multiple agents properly in the workload summary
        raise NotImplementedError

    def toJSON(self):
        """
        _toJSON_

        Bin the histogram if any, calculate the standard deviation. Store
        the internal data needed for reconstruction of the histogram
        from JSON and call superclass toJSON method.
        """
        if self.nPoints:
            self.stdDev = calculateStdDevFromQ(self.QValue, self.nPoints)
        if not self.binned and self.storeHistogram:
            self.binHistogram()
        self.jsonInternal = {}
        self.jsonInternal['yLabel'] = self.yLabel
        self.jsonInternal['QValue'] = self.QValue
        self.jsonInternal['nPoints'] = self.nPoints
        return SummaryHistogram.toJSON(self)

    def binHistogram(self):
        """
        _binHistogram_

        Histograms of continuous data must be binned,
        this takes care of that using given or optimal parameters.
        Note that this modifies the data object,
        and points can't be added to the histogram after this.
        """

        if not self.nPoints:
            return

        self.binned = True

        # Number of bins can be specified or calculated based on number of points
        nBins = self.fixedNBins
        if nBins is None:
            nBins = int(math.floor((5.0 / 3.0) * math.pow(self.nPoints, 1.0 / 3.0)))

        # Define min and max
        if not self.dropOutliers:
            upperLimit = max(self.data.keys())
            lowerLimit = min(self.data.keys())
        else:
            stdDev = calculateStdDevFromQ(self.QValue, self.nPoints)
            upperLimit = self.average + (stdDev * self.sigmaLimit)
            lowerLimit = self.average - (stdDev * self.sigmaLimit)

        # Incremental delta
        delta = abs(float(upperLimit - lowerLimit)) / nBins

        # Build the bins, it's a list of tuples for now
        bins = []

        a = lowerLimit
        b = lowerLimit + delta
        while len(bins) < nBins:
            bins.append((a, b))
            a += delta
            b += delta
        # Go through data and populate the binned histogram
        binnedHisto = {}
        currentBin = 0
        currentPoint = 0
        sortedData = sorted(self.data.keys())
        while currentPoint < len(sortedData):
            point = sortedData[currentPoint]
            encodedTuple = "%s,%s" % (bins[currentBin][0], bins[currentBin][1])
            if encodedTuple not in binnedHisto:
                binnedHisto[encodedTuple] = 0
            if point > upperLimit or point < lowerLimit:
                currentPoint += 1
            elif currentBin == len(bins) - 1:
                binnedHisto[encodedTuple] += self.data[point]
                currentPoint += 1
            elif point >= bins[currentBin][0] and point < bins[currentBin][1]:
                binnedHisto[encodedTuple] += self.data[point]
                currentPoint += 1
            else:
                currentBin += 1

        self.data = binnedHisto
        return



#!/usr/bin/env python
"""
File       : adjustMongoDocs.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: adjust documents in MongoDB with given schema
The script loops over given db documents (from given db name
and collection) and add to each document a new schema provided
via input file name, e.g.

# new schema addition:
cat > /tmp/schema.json << EOF
{"foo": 1}
EOF

# run script to update all documents in test.test database and add to them
# a json structure presented in /tmp/schema.json file
adjustMongoDocs.py --fin=/tmp/schema.json --dburi=localhost:8230 --dbname=test --dbcoll=test --verbose 1

NOTE: the script can run from any location assuming it has proper db uri with
credentials and availability of mongo db outside of its own cluster, but most
likely it will be run on mongo db k8s cluster
"""

import argparse
import json
from pymongo import MongoClient


class OptionParser(object):
    "Helper class to handle script options"

    def __init__(self):
        "User based option parser"
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument("--fin", action="store",
                                 dest="fin", default="", help="Input file with update fields")
        self.parser.add_argument("--dburi", action="store",
                                 dest="dburi", default="", help="MongoDB URI")
        self.parser.add_argument("--dbname", action="store",
                                 dest="dbname", default="", help="MongoDB database")
        self.parser.add_argument("--dbcoll", action="store",
                                 dest="dbcoll", default="", help="MongoDB collection")
        self.parser.add_argument("--verbose", action="store",
                                 dest="verbose", default=0, help="verbosity level")


def adjust():
    """
    helper script to update documents in MongoDB with given db name/collection
    and schema file (which should contain new addition, in form of JSON, to the
    db documents)
    """
    optmgr = OptionParser()
    opts = optmgr.parser.parse_args()
    conn = MongoClient(host=opts.dburi)
    dbname = opts.dbname
    dbcoll = opts.dbcoll
    verbose = opts.verbose
    schema = {}
    with open(opts.fin, 'r', encoding='utf-8') as istream:
        schema = json.load(istream)
    if schema:
        if verbose:
            print("read data from '%s', %s/%s" % (opts.dburi, dbname, dbcoll))
        records = [r for r in conn[dbname][dbcoll].find()]
        for doc in records:
            if verbose:
                print("update doc=%s with schema=%s" % (doc, schema))
            conn[dbname][dbcoll].update_one(doc, {'$set': schema}, upsert=True)


if __name__ == '__main__':
    adjust()



"""
File       : MSTransferor.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
             Alan Malta <alan dot malta AT cern dot ch >
Description: MSTransferor class provides the whole logic for
central production workflow's input data placement.

This is NOT a thread-safe module, even though some internal
tasks might be extended to multi-threading in the future.
"""
# futures
from __future__ import division, print_function
from future.utils import listvalues, listitems
from future import standard_library
standard_library.install_aliases()

# system modules
from operator import itemgetter
from pprint import pformat
from retry import retry
from copy import deepcopy

# WMCore modules
from Utils.IteratorTools import grouper
from WMCore.MicroService.DataStructs.DefaultStructs import TRANSFEROR_REPORT,\
    TRANSFER_RECORD, TRANSFER_COUCH_DOC
from WMCore.MicroService.Tools.Common import (teraBytes, isRelVal, getPileupDocs)
from WMCore.MicroService.MSCore.MSCore import MSCore
from WMCore.MicroService.MSTransferor.RequestInfo import RequestInfo
from WMCore.MicroService.MSTransferor.DataStructs.RSEQuotas import RSEQuotas
from WMCore.Services.CRIC.CRIC import CRIC
from WMCore.Services.Rucio.RucioUtils import GROUPING_ALL


def newTransferRec(dataIn):
    """
    Create a basic transfer record to be appended to a transfer document
    :param dataIn: dictionary with information relevant to this transfer doc
    :return: a transfer record dictionary
    """
    record = deepcopy(TRANSFER_RECORD)
    record["dataset"] = dataIn['name']
    record["dataType"] = dataIn['type']
    record["campaignName"] = dataIn['campaign']
    return record


def newTransferDoc(reqName, transferRecords):
    """
    Create a transfer document which is meant to be created in
    central CouchDB
    :param reqName: string with the workflow name
    :param transferRecords: list of dictionaries with transfer records
    :return: a transfer document dictionary
    """
    doc = dict(TRANSFER_COUCH_DOC)
    doc["workflowName"] = reqName
    doc["transfers"] = transferRecords
    return doc

class MSTransferor(MSCore):
    """
    MSTransferor class provide whole logic behind
    the transferor module.
    """

    def __init__(self, msConfig, logger=None):
        """
        Runs the basic setup and initialization for the MS Transferor module
        :param microConfig: microservice configuration
        """
        super(MSTransferor, self).__init__(msConfig, logger=logger)

        # minimum percentage completion for dataset/blocks subscribed
        self.msConfig.setdefault("minPercentCompletion", 99)
        # minimum available storage to consider a resource good for receiving data
        self.msConfig.setdefault("minimumThreshold", 1 * (1000 ** 4))  # 1TB
        # limit MSTransferor to this amount of requests per cycle
        self.msConfig.setdefault("limitRequestsPerCycle", 500)
        # Send warning messages for any data transfer above this threshold.
        # Set to negative to ignore.
        self.msConfig.setdefault("warningTransferThreshold", 100. * (1000 ** 4))  # 100TB
        # weight expression for the input replication rules
        self.msConfig.setdefault("rucioRuleWeight", 'ddm_quota')
        # Workflows with open running timeout are used for growing input dataset, thus
        # make a container level rule for the whole container whenever the open running
        # timeout is larger than what is configured (or the default of 7 days below)
        self.msConfig.setdefault("openRunning", 7 * 24 * 60 * 60)
        # define the pileup query to be executed through MSPileup
        self.pileupQuery = self.msConfig.get("pileupQuery",
                                             {"query": {"active": True}, "filters": ["expectedRSEs", "pileupName"]})

        quotaAccount = self.msConfig["rucioAccount"]

        self.rseQuotas = RSEQuotas(quotaAccount, self.msConfig["quotaUsage"],
                                   minimumThreshold=self.msConfig["minimumThreshold"],
                                   verbose=self.msConfig['verbose'], logger=logger)
        self.reqInfo = RequestInfo(self.msConfig, self.rucio, self.logger)

        self.cric = CRIC(logger=self.logger)
        self.pileupDocs = []
        self.campaigns = {}
        self.psn2pnnMap = {}
        self.pnn2psnMap = {}
        self.dsetCounter = 0
        self.blockCounter = 0
        # service name used to route alerts via AlertManager
        self.alertServiceName = "ms-transferor"

    @retry(tries=3, delay=2, jitter=2)
    def updateCaches(self):
        """
        Fetch some data required for the transferor logic, e.g.:
         * account limits from Rucio
         * account usage from Rucio
         * unified configuration
         * all campaign configuration
         * PSN to PNN map from CRIC
        """
        self.logger.info("Updating RSE/PNN quota and usage")
        self.rseQuotas.fetchStorageQuota(self.rucio)
        self.rseQuotas.fetchStorageUsage(self.rucio)
        self.rseQuotas.evaluateQuotaExceeded()
        if not self.rseQuotas.getNodeUsage():
            raise RuntimeWarning("Failed to fetch storage usage stats")

        self.logger.info("Updating all local caches...")
        self.dsetCounter = 0
        self.blockCounter = 0
        self.pileupDocs = getPileupDocs(self.msConfig['mspileupUrl'], self.pileupQuery)
        campaigns = self.reqmgrAux.getCampaignConfig("ALL_DOCS")
        self.psn2pnnMap = self.cric.PSNtoPNNMap()
        self.pnn2psnMap = self.cric.PNNtoPSNMap()
        if not campaigns:
            raise RuntimeWarning("Failed to fetch the campaign configurations")
        elif not self.psn2pnnMap:
            raise RuntimeWarning("Failed to fetch PSN x PNN map from CRIC")
        else:
            # let's make campaign look-up easier and more efficient
            self.campaigns = {}
            for camp in campaigns:
                self.campaigns[camp['CampaignName']] = camp
        self.rseQuotas.printQuotaSummary()

    def execute(self, reqStatus):
        """
        Executes the whole transferor logic
        :param reqStatus: request status to process
        :return:
        """
        counterWorkflows = 0
        counterFailedRequests = 0
        counterProblematicRequests = 0
        counterSuccessRequests = 0
        summary = dict(TRANSFEROR_REPORT)
        self.logger.info("Service set to process up to %s requests per cycle.",
                         self.msConfig["limitRequestsPerCycle"])
        try:
            requestRecords = self.getRequestRecords(reqStatus)
            self.updateReportDict(summary, "total_num_requests", len(requestRecords))
            self.logger.info("Retrieved %s requests.", len(requestRecords))
        except Exception as err:  # general error
            requestRecords = []
            msg = "Unknown exception while fetching requests from ReqMgr2. Error: %s", str(err)
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)

        try:
            self.updateCaches()
            self.updateReportDict(summary, "total_num_active_pileups", len(self.pileupDocs))
            self.updateReportDict(summary, "total_num_campaigns", len(self.campaigns))
            self.updateReportDict(summary, "nodes_out_of_space", list(self.rseQuotas.getOutOfSpaceRSEs()))
        except RuntimeWarning as ex:
            msg = "All retries exhausted! Last error was: '%s'" % str(ex)
            msg += "\nRetrying to update caches again in the next cycle."
            self.logger.error(msg)
            self.updateReportDict(summary, "error", msg)
            return summary
        except Exception as ex:
            msg = "Unknown exception updating caches. Error: %s" % str(ex)
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)
            return summary

        # process all requests
        for reqSlice in grouper(requestRecords, 100):
            self.logger.info("Processing workflows from %d to %d.",
                             counterWorkflows + 1, counterWorkflows + len(reqSlice))
            # execute data discovery
            reqResults = self.reqInfo(reqSlice, self.pileupDocs)
            self.logger.info("%d requests information completely processed.", len(reqResults))

            for wflow in reqResults:
                if not self.verifyCampaignExist(wflow):
                    counterProblematicRequests += 1
                    continue

                if not self.passSecondaryCheck(wflow):
                    self.alertPUMisconfig(wflow.getName())
                    counterProblematicRequests += 1
                    continue

                # find accepted RSEs for the workflow
                rseList = self.getAcceptedRSEs(wflow)

                # now check where input primary and parent blocks will need to go
                self.checkDataLocation(wflow, rseList)

                try:
                    success, transfers = self.makeTransferRequest(wflow, rseList)
                except Exception as ex:
                    success = False
                    self.alertUnknownTransferError(wflow.getName())
                    msg = "Unknown exception while making transfer request for %s " % wflow.getName()
                    msg = "\tError: %s" % str(ex)
                    self.logger.exception(msg)
                if success:
                    # then create a document in ReqMgr Aux DB
                    self.logger.info("Transfer requests successful for %s. Summary: %s",
                                     wflow.getName(), pformat(transfers))
                    if self.createTransferDoc(wflow.getName(), transfers):
                        self.logger.info("Transfer document successfully created in CouchDB for: %s", wflow.getName())
                        # then move this request to staging status
                        self.change(wflow.getName(), 'staging', self.__class__.__name__)
                        counterSuccessRequests += 1
                    else:
                        counterFailedRequests += 1
                        self.alertTransferCouchDBError(wflow.getName())
                else:
                    counterFailedRequests += 1
            # it can go slightly beyond the limit. It's evaluated for every slice
            if counterSuccessRequests >= self.msConfig["limitRequestsPerCycle"]:
                msg = "Transferor succeeded acting on %d workflows in this cycle. " % counterSuccessRequests
                msg += "Which exceeds the configuration limit set to: %s" % self.msConfig["limitRequestsPerCycle"]
                self.logger.info(msg)
                break
            counterWorkflows += len(reqSlice)

        self.logger.info("Summary for this cycle is:")
        self.logger.info("    * there were %d problematic requests;", counterProblematicRequests)
        self.logger.info("    * there were %d failed requests;", counterFailedRequests)
        self.logger.info("    * there were %d successful requests;", counterSuccessRequests)
        self.logger.info("    * a total of %d datasets were subscribed;", self.dsetCounter)
        self.logger.info("    * a total of %d blocks were subscribed.", self.blockCounter)
        self.updateReportDict(summary, "success_request_transition", counterSuccessRequests)
        self.updateReportDict(summary, "failed_request_transition", counterFailedRequests)
        self.updateReportDict(summary, "problematic_requests", counterProblematicRequests)
        self.updateReportDict(summary, "num_datasets_subscribed", self.dsetCounter)
        self.updateReportDict(summary, "num_blocks_subscribed", self.blockCounter)
        self.updateReportDict(summary, "nodes_out_of_space", list(self.rseQuotas.getOutOfSpaceRSEs()))
        return summary

    def getRequestRecords(self, reqStatus):
        """
        Queries ReqMgr2 for requests in a given status, sort them by priority
        and return a subset of each request with important information for the
        data placement algorithm.
        """
        self.logger.info("Fetching requests in status: %s", reqStatus)
        # get requests from ReqMgr2 data-service for given status
        reqData = self.reqmgr2.getRequestByStatus([reqStatus], detail=True)

        # we need to first put these requests in order of priority, as done for GQ...
        orderedRequests = []
        for requests in reqData:
            orderedRequests = listvalues(requests)
        orderedRequests.sort(key=itemgetter('RequestPriority'), reverse=True)

        return orderedRequests

    def verifyCampaignExist(self, wflow):
        """
        Check whether there is a campaign for the primary dataset.
        :param wflow: a workflow object
        :return: True if campaigns exist, False otherwise
        """
        for dataIn in wflow.getDataCampaignMap():
            if dataIn["type"] == "primary":
                if dataIn['campaign'] not in self.campaigns:
                    msg = "Workflow: %s has to transfer dataset: %s under the campaign: %s. "
                    msg += "This campaign does not exist and needs to be created. Skipping this workflow!"
                    self.logger.warning(msg, wflow.getName(), dataIn['name'], dataIn['campaign'])
                    return False
        return True

    def passSecondaryCheck(self, wflow):
        """
        Check if the workflow uses active pileup and with valid location.
        :param wflow: workflow object
        :return: boolean whether the workflow is good to go or not
        """
        pileupInput = wflow.getSecondarySummary()
        if not pileupInput:
            # nothing to be done here
            return True
        for puName, puData in pileupInput.items():
            if puData['locations'] == []:
                msg = f"Workflow {wflow.getName()} requires pileup dataset {puName} "
                msg += "which is either not active or does not exist in MSPileup."
                self.logger.warning(msg)
                return False
        return True

    def getAcceptedRSEs(self, wflow):
        """
        Given a workflow object, find it's final accepted list of
        RSEs for input data placement. This is based on:
         * TrustSitelists and TrustPUSitelists;
         * Workflow SiteWhitelist and SiteBlacklist
         * Pileup location(s)
        Note that it does NOT account for RSEs out of quota.
        :param wflow: wflow object
        :return: a list with unique RSE names
        """
        # workflow level site lists. In case there is no pileup or SecAAA=True
        wflowRSEs = self._getPNNsFromPSNs(wflow.getSitelist())
        if wflow.getPileupDatasets() and not wflow.getReqParam("TrustPUSitelists"):
            # otherwise, data needs to be placed where pileup is
            wflowRSEs = wflowRSEs & wflow.getPURSElist()

        return list(wflowRSEs)

    def checkDataLocation(self, wflow, rseList):
        """
        Check which data is already in place (according to the site lists
        and pileup data location) and remove them from the data placement
        if already available anywhere.
        If workflow has XRootD/AAA enabled, data location can be outside of
        the SiteWhitelist.
        :param wflow: workflow object
        :param rseList: list of RSE names allowed for a given workflow
        :return: None
        """
        if not wflow.getInputDataset():
            return

        primAAA = wflow.getReqParam("TrustSitelists")
        secAAA = wflow.getReqParam("TrustPUSitelists")
        msg = f"Checking data location for request: {wflow.getName()}, "
        msg += f"TrustSitelists: {primAAA}, TrustPUSitelists: {secAAA}, "
        msg += f"and final accepted RSEs: {rseList}"
        self.logger.info(msg)

        for methodName in ("getPrimaryBlocks", "getParentBlocks"):
            inputBlocks = getattr(wflow, methodName)()
            self.logger.info("Request %s has %d initial blocks from %s",
                             wflow.getName(), len(inputBlocks), methodName)

            for block, blockDict in listitems(inputBlocks):  # dict can change size here
                blockLocation = self._diskPNNs(blockDict['locations'])
                if not blockLocation:
                    self.logger.info("Primary/parent block %s not available in any disk storage", block)
                elif primAAA:
                    msg = "Primary/parent block %s already in place (via AAA): %s" % (block, blockLocation)
                    self.logger.info(msg)
                    inputBlocks.pop(block)
                else:
                    commonLocation = set(blockLocation) & rseList
                    if commonLocation:
                        self.logger.info("Primary/parent block %s already in place: %s", block, commonLocation)
                        inputBlocks.pop(block)
                    else:
                        self.logger.info("block: %s will need data placement!!!", block)

            self.logger.info("Request %s has %d final blocks from %s",
                             wflow.getName(), len(getattr(wflow, methodName)()), methodName)


    def makeTransferRequest(self, wflow, rseList):
        """
        Checks which input data has to be transferred, select the final destination if needed,
        create the transfer record to be stored in Couch, and create the DM placement request.
        This method does the following:
          1. return if there is no workflow data to be transferred
          2. check if the data input campaign is in the database, skip if not
          3. _getValidSites: using the workflow site lists and the campaign configuration,
             find a common list of sites (converted to PNNs). If the PNN is out of quota,
             it's also removed from this list
          4. create the transfer record dictionary
          5. for every final node
             5.1. if it's a pileup dataset, pick a random node and subscribe the whole container
             5.2. else, retrieve chunks of blocks to be subscribed (evenly distributed)
             5.3. update node usage with the amount of data subscribed
          6. re-evaluate nodes with quota exceeded
          7. return the transfer record, with a list of transfer IDs
        :param wflow: workflow object
        :param rseList: list of RSE names allowed for a given workflow
        :return: boolean whether it succeeded or not, and a list of transfer records
        """
        response = []
        success = True
        if not (wflow.getParentBlocks() or wflow.getPrimaryBlocks()):
            self.logger.info("Request %s does not have any further data to transfer", wflow.getName())
            return success, response

        self.logger.info("Handling data subscriptions for request: %s", wflow.getName())

        for dataIn in wflow.getDataCampaignMap():
            if dataIn["type"] == "parent":
                msg = "Skipping 'parent' data placement (done with the 'primary' data), for: %s" % dataIn
                self.logger.info(msg)
                continue
            elif dataIn["type"] == "secondary":
                # already performed by MSPileup
                continue

            if not isRelVal(wflow.data):
                # enforce RSE quota
                rses = list(set(rseList) & self.rseQuotas.getAvailableRSEs())
            else:
                rses = rseList

            if not rses:
                msg = f"Workflow: {wflow.getName()} could have data placed at: {rseList}, "
                msg += "but those are all out of quota. Skipping it till next cycle"
                self.logger.warning(msg)
                return False, response

            # create a transfer record data structure
            transRec = newTransferRec(dataIn)
            # figure out dids, number of copies and which grouping to use
            dids, didsSize = wflow.getInputData()
            grouping = wflow.getRucioGrouping()
            copies = wflow.getReplicaCopies()
            if not dids:
                # no valid files in any blocks, it will likely fail in global workqueue
                self.logger.warning("  found 0 primary/parent blocks for dataset: %s, moving on...", dataIn['name'])
                return success, response

            success, transferId = self.makeTransferRucio(wflow, dataIn, dids, didsSize,
                                                         grouping, copies, rses)
            if not success:
                # stop any other data placement for this workflow
                msg = "There were failures transferring data for workflow: %s. Will retry again later."
                self.logger.warning(msg, wflow.getName())
                break
            if transferId:
                if isinstance(transferId, (set, list)):
                    transRec['transferIDs'].update(transferId)
                else:
                    transRec['transferIDs'].add(transferId)

            # and update some instance caches
            if grouping == GROUPING_ALL:
                self.dsetCounter += 1
            else:
                self.blockCounter += len(dids)

        transRec['transferIDs'] = list(transRec['transferIDs'])
        response.append(transRec)

        return success, response

    def makeTransferRucio(self, wflow, dataIn, dids, dataSize, grouping, copies, nodes):
        """
        Creates a Rucio replication rule

        :param wflow: the workflow object
        :param dataIn: short summary of the data to be placed
        :param dids: a list of the DIDs to be added to the rule
        :param dataSize: amount of data being placed by this rule
        :param grouping: whether blocks need to be placed altogether (ALL)
                         or if the can be scattered around (DATASET).
        :param copies: integer with the number of copies to use in the rule
        :param nodes: list of nodes/RSE
        :return: a boolean flagging whether it succeeded or not, and the rule id
        """
        success, transferId = True, set()

        ruleAttrs = {'copies': copies,
                     'activity': 'Production Input',
                     'lifetime': self.msConfig['rulesLifetime'],
                     'account': self.msConfig['rucioAccount'],
                     'grouping': grouping,
                     'weight': self.msConfig['rucioRuleWeight'],
                     'meta': {'workflow_group': wflow.getWorkflowGroup()},
                     'comment': 'WMCore MSTransferor input data placement'}

        rseExpr = "|".join(nodes)

        if self.msConfig.get('enableDataTransfer', True):
            # Force request-only subscription
            # to any data transfer going above some threshold (do not auto-approve)
            aboveWarningThreshold = (self.msConfig.get('warningTransferThreshold') > 0. and
                                     dataSize > self.msConfig.get('warningTransferThreshold'))

            # Then make the data subscription, for real!!!
            self.logger.info("Creating rule for workflow %s with %d DIDs in container %s, RSEs: %s, grouping: %s",
                             wflow.getName(), len(dids), dataIn['name'], rseExpr, grouping)
            try:
                res = self.rucio.createReplicationRule(dids, rseExpr, **ruleAttrs)
            except Exception as exc:
                msg = "Hit a bad exception while creating replication rules for DID: %s. Error: %s"
                self.logger.error(msg, dids, str(exc))
                success = False
            else:
                if res:
                    # it could be that some of the DIDs already had such rule in
                    # place, so we might be retrieving a bunch of rule ids instead of
                    # a single one
                    self.logger.info("Rules successful created for %s : %s", dataIn['name'], res)
                    transferId.update(res)
                    # send an alert, if needed
                    self.alertLargeInputData(aboveWarningThreshold, transferId, wflow.getName(), dataSize, dataIn)
                else:
                    self.logger.error("Failed to create rule for %s, will retry later", dids)
                    success = False
        else:
            msg = "DRY-RUN: making Rucio rule for workflow: %s, dids: %s, rse: %s, kwargs: %s"
            self.logger.info(msg, wflow.getName(), dids, rseExpr, ruleAttrs)
        return success, transferId

    def alertPUMisconfig(self, workflowName):
        """
        Send alert to Prometheus with PU misconfiguration error
        """
        alertName = "{}: PU misconfiguration error. Workflow: {}".format(self.alertServiceName,
                                                                         workflowName)
        alertSeverity = "high"
        alertSummary = "[MSTransferor] Workflow cannot proceed due to some PU misconfiguration."
        alertDescription = "Workflow: {} could not proceed due to some PU misconfiguration,".format(workflowName)
        alertDescription += "so it will be skipped."
        self.sendAlert(alertName, alertSeverity, alertSummary, alertDescription,
                       self.alertServiceName)
        self.logger.critical(alertDescription)

    def alertUnknownTransferError(self, workflowName):
        """
        Send alert to Prometheus with unknown transfer error
        """
        alertName = "{}: Transfer request error. Workflow: {}".format(self.alertServiceName,
                                                                         workflowName)
        alertSeverity = "high"
        alertSummary = "[MSTransferor] Unknown exception while making transfer request."
        alertDescription = "Unknown exception while making Transfer request for workflow: {}".format(workflowName)
        self.sendAlert(alertName, alertSeverity, alertSummary, alertDescription,
                       self.alertServiceName)

    def alertTransferCouchDBError(self, workflowName):
        """
        Send alert to Prometheus with CouchDB transfer error
        """
        alertName = "{}: Failed to create a transfer document in CouchDB for workflow: {}".format(self.alertServiceName,
                                                                         workflowName)
        alertSeverity = "high"
        alertSummary = "[MSTransferor] Transfer document could not be created in CouchDB."
        alertDescription = "Workflow: {}, failed request  due to error posting to CouchDB".format(workflowName)
        self.sendAlert(alertName, alertSeverity, alertSummary, alertDescription,
                       self.alertServiceName)
        self.logger.warning(alertDescription)


    def alertLargeInputData(self, aboveWarningThreshold, transferId, wflowName, dataSize, dataIn):
        """
        Evaluates whether the amount of data placed is too big, if so, send an alert
        notification to a few persons
        :param aboveWarningThreshold: boolean flag saying if the thresholds was exceeded or not
        :param transferId: rule/transfer request id
        :param wflowName: name of the workflow
        :param dataSize: total amount of data subscribed
        :param dataIn: short summary of the workflow data
        """
        # Warn about data transfer subscriptions going above some threshold
        if aboveWarningThreshold:
            alertName = "{}: input data transfer over threshold: {}".format(self.alertServiceName,
                                                                            wflowName)
            alertSeverity = "high"
            alertSummary = "[MS] Large pending data transfer under request id: {}".format(transferId)
            alertDescription = "Workflow: {} has a large amount of ".format(wflowName)
            alertDescription += "data subscribed: {} TB, ".format(teraBytes(dataSize))
            alertDescription += "for {} data: {}.""".format(dataIn['type'], dataIn['name'])

            self.sendAlert(alertName, alertSeverity, alertSummary, alertDescription,
                           self.alertServiceName)
            self.logger.warning(alertDescription)

    def _getValidSites(self, wflow, dataIn):
        """
        Given a workflow object and the data short summary, find out
        the Campaign name, the workflow SiteWhitelist, map the PSNs to
        PNNs and finally remove PNNs without space
        can still receive data
        :param wflow: the workflow object
        :param dataIn: short summary of data to be transferred
        :return: a unique and ordered list of PNNs to take data
        """
        campConfig = self.campaigns[dataIn['campaign']]
        psns = wflow.getSitelist()

        if dataIn["type"] == "primary":
            if campConfig['SiteWhiteList']:
                psns = set(psns) & set(campConfig['SiteWhiteList'])
            if campConfig['SiteBlackList']:
                psns = set(psns) - set(campConfig['SiteBlackList'])

        self.logger.info("  final list of PSNs to be use: %s", psns)
        pnns = self._getPNNsFromPSNs(psns)

        if isRelVal(wflow.data):
            self.logger.info("RelVal workflow '%s' ignores sites out of quota", wflow.getName())
            return list(pnns)

        self.logger.info("List of out-of-space RSEs dropped for '%s' is: %s",
                         wflow.getName(), pnns & self.rseQuotas.getOutOfSpaceRSEs())
        return list(pnns & self.rseQuotas.getAvailableRSEs())

    def createTransferDoc(self, reqName, transferRecords):
        """
        Enrich the records returned from the data placement logic, wrap them up
        in a single document and post it to CouchDB
        :param reqName: the workflow name
        :param transferRecords: list of dictionaries records, or empty if no input at all
        :return: True if operation is successful, else False
        """
        doc = newTransferDoc(reqName, transferRecords)
        # Use the update/put method, otherwise it will fail if the document already exists
        if self.reqmgrAux.updateTransferInfo(reqName, doc):
            return True
        self.logger.error("Failed to create transfer document in CouchDB. Will retry again later.")
        return False

    def _getPNNsFromPSNs(self, psnList):
        """
        Given a list/set of PSNs, return a set of valid PNNs.
        Note that T3, Tape and a few other PNNs are never returned.
        """
        pnns = set()
        for psn in psnList:
            for pnn in self.psn2pnnMap.get(psn, []):
                if pnn == "T2_CH_CERNBOX" or pnn.startswith("T3_"):
                    pass
                elif pnn.endswith("_Tape") or pnn.endswith("_MSS") or pnn.endswith("_Export"):
                    pass
                else:
                    pnns.add(pnn)
        return pnns

    def _getPSNsFromPNNs(self, pnnList):
        """
        Given a list/set of PNNs, return a set of valid PSNs.
        Note that T3 sites are never returned.
        """
        psns = set()
        for pnn in pnnList:
            for psn in self.pnn2psnMap.get(pnn, []):
                if psn.startswith("T3_"):
                    pass
                else:
                    psns.add(psn)
        return psns

    def _diskPNNs(self, pnnList):
        """
        Provided a list of PNN locations, return another list of
        PNNs without mass storage and T3 sites
        :param pnnList: list of PNN strings
        :return: a set of strings with filtered out PNNs
        """
        diskPNNs = set()
        for pnn in pnnList:
            if pnn == "T2_CH_CERNBOX" or pnn.startswith("T3_"):
                pass
            elif pnn.endswith("_Tape") or pnn.endswith("_MSS") or pnn.endswith("_Export"):
                pass
            else:
                diskPNNs.add(pnn)
        return diskPNNs
